
--
-- Configuraes para compatibilizar script tanto com  averso 4.0 quanto
-- com a verso 5.1 da biblioteca Lua
--
-- Para que este arquivo poss manter-se nico para as duas verses, deve 
-- ser tomado cuidado para que nao seja utilizada nenhuma caracterstica 
-- da verso 5.1 que gere um erro de sintaxe na verso 4.0 e vice versa
-- (exs: operador % para upvalues na verso 4, operador # na verso 5.1, ...)
--
if _VERSION == 'Lua 5.1' then
  setglobal = function(name, val) _G[name] = val end
  globals   = function() return _G end
  floor     = math.floor
  mod       = math.fmod
  tinsert   = function(t, v) 
                -- Aceita nil e nao usa operador # (erro de sitaxe em Lua 4.0)
                local pos = t.n or table.maxn(t)
                t[pos+1] = v
                t.n = pos + 1
              end
  tremove   = function(t, pos)
                local size = t.n or table.maxn(t)
                table.remove(t, pos)
                t.n = size - 1
              end
  getn      = function(t) 
                if not t.n then 
                  t.n = table.maxn(t) 
                end
                return t.n
              end
  openfile  = io.open
  closefile = io.close
  write     = function (f, ...) return f:write(unpack(arg)) end
  strrep    = string.rep 
  format    = string.format
  gsub      = string.gsub
  strfind   = string.find
  dostring  = function(s, n) assert(loadstring(s, n))() end
  loadtemplate = dofile
else
  loadtemplate = function(f) assert(dofile(f)) end
end


-- Tabela auxiliar contendo o nome dos campos utilizados nas tabelas
-- que descrevem os eixos de uma rea grfica
CfgGraphAreaFields = {index = {'xaxeIndex', 'y1axeIndex', 'y2axeIndex', 'y3axeIndex'},
                      title = {'xaxeTitle', 'y1axeTitle', 'y2axeTitle', 'y3axeTitle'},
                      series = {'', 'y1axeSeries', 'y2axeSeries', 'y3axeSeries'},
                     }

CfgSeriesColors     = {red, green, blue, yellow, magenta, cyan, gray}
CfgSeriesDarkColors = {dark_red, dark_green, dark_blue, dark_yellow, 
                       dark_magenta, dark_cyan, dark_gray}
CfgSeriesMarkStyles = {MS_SQUARE + MS_SOLID_SQUARE,     MS_CIRCLE + MS_SOLID_CIRCLE,    
                       MS_TRIANGLE + MS_SOLID_TRIANGLE, MS_DIAMOND + MS_SOLID_DIAMOND,   
                       MS_PLUS, MS_X }

-- Cria um novo nome de objeto, nico, a partir de um prefixo e um contador 
-- de ocorrencias daquele prefixo
CfgPrefixTable = {}
function CfgBuildName(prefix)
  local n = CfgPrefixTable[prefix] or 0
  n = n + 1
  CfgPrefixTable[prefix] = n
  return prefix..tostring(n)
end

-- Seta tag do eixo de acordo com a codificacao:
--   areatag * 1000 + axetag * 100 
CfgTagTable = {}
function CfgSetAxeTag(axe, areatag, axetag)
  assert(axetag  >= 0 and axetag <= 9)
  assert(areatag >= 1 and areatag <= 99)
  local tag = areatag * 1000 + axetag * 100
  assert(CfgTagTable[tag] == nil)
  CfgTagTable[tag] = 0
  axe:setUserTag(tag)
end

-- Seta tag do objeto de acordo com a codificacao:
--   areatag * 1000 + axetag * 100 + objecttag
function CfgSetObjectTag(obj, areatag, axetag)
  assert(axetag  >= 0 and axetag <= 9)
  assert(areatag >= 1 and areatag <= 99)
  local tag = areatag * 1000 + axetag * 100
  local n = CfgTagTable[tag] or 0
  assert(n < 99)
  n = n + 1
  CfgTagTable[tag] = n
  obj:setUserTag(tag + n)
end


-- Retorna o tag da area associado ao objeto recebido como parametro
function CfgGetAreaTag(obj)
  return floor(obj:userTag() / 1000)
end

-- Retorna o tag do eixo associado ao objeto recebido como parametro
function CfgGetAxeTag(obj)
  return floor(mod(obj:userTag(), 1000) / 100)
end


-- Cria uma nova PlotArea dentro dos limites e associada aos eixos recebidos
-- como parmetro. Configra a rea, adiciona grids e seta identificao padro
-- da rea.
--
-- area_data : Tabela contendo os dados da rea grfica.  Utiliza o campo tag para
--             gerar a numerao padro das reas (codificada como AEOO, ou seja,
--             reas esto codificadas no milhar do tag, o eixo nas centenas e os
--             objetos associados na unidade e dezenas)
-- leftgl, 
-- rightgl,
-- bottomgl,
-- topgl     : Limites da rea grfica
-- xaxe,
-- yaxe      : Eixos associados  rea grfica
-- gridlist  : Lista preenchida com os grids criados para a rea
--             se nil, grids no so criados
--
-- Retorna a PlotArea criada
--
function CfgBuildPlotArea(area_data, leftgl, rightgl, bottomgl, topgl, 
                          xaxe, yaxe, gridlist)

  local name = CfgBuildName('plotarea_')
  local area = SgPlotArea:new(CurrentGraphic, name, leftgl, rightgl, 
                              bottomgl, topgl, xaxe, yaxe)

  area:setNameExt(area_data.title)
  assert(area_data.tag > 0)
  area:setUserTag(area_data.tag * 1000)

  CfgSetPlotAreaStyle(area)
  if gridlist then 
    CfgAddGrids(area, xaxe, yaxe, gridlist)
  end

  setglobal(name, area)
  return area
end

--
-- Define as caracteristicas de estilo da serie de uma forma padro
--
-- serie : Serie que ter as caracteristicas definidas
-- plot : Tipo da plotagem dos dados
function CfgSetDefaultSeriesStyle(serie,plot)
  local color_index = CfgSeriesColors.index or 1
  
  local color     = CfgSeriesColors[color_index]
  local darkcolor = CfgSeriesDarkColors[color_index]
  color_index = color_index + 1
  if color_index > getn(CfgSeriesColors) then color_index = 1 end
  CfgSeriesColors.index = color_index
  
  serie:style():pen():setColor(color)

  if plot == 'PointSeries' or plot == 'LineSeries' then
  
	local style_index = CfgSeriesMarkStyles.index or 1
	local style       = CfgSeriesMarkStyles[style_index] or 1

	style_index = style_index + 1
	if style_index > getn(CfgSeriesMarkStyles) then style_index = 1 end
	CfgSeriesMarkStyles.index = style_index

	serie:markerStyle():setType(style)
	serie:markerStyle():setSize(2.0)
	serie:markerStyle():pen():setColor(darkcolor)
	serie:markerStyle():brush():setFgColor(color)
  end
  
  if plot == 'HistogramSeries' then
	serie:setFillAxe (1,0.0);
	serie:style():brush():setFgColor(darkcolor)
  end
  
  if plot == 'PieSeries' or plot == 'RoseSeries' then
	serie:style():pen():setColor(SgColor:new(0,0,0))
  end
  
end

--
-- Cria uma nova serie para representar o data_src recebido como 
-- parametro.  Se a curva possuir uma funcao de inicializacao 
-- configurada, utiliza esta funcao para inicializar a serie.
-- Caso contrario inicializa a mesma segundo uma tabela propria 
-- para fazer com que cada serie criada tenha uma representacao
-- distinta
--
-- data_src  : Fonte de dados
-- cur_index : Indice do data_src em SeriesList
-- title     : Titulo da srie a ser apresentada na legenda.  Se nil, 
--             utiliza o valor padro configurado em SeriesList
-- order : ordem da serie quando for multidata
--         
function CfgBuildSeries(data_src, cur_index, title, order)
  local name = CfgBuildName('series'..SeriesList[cur_index].id..'_')

  local ser;
  local plot = SeriesList[cur_index].plotType or 'LineSeries';
  	
  if plot == 'PointSeries' then
	ser = SgPointSeries:new(name, data_src, true)
  end
  if plot == 'LineSeries' then
	ser = SgLineSeries:new(name, data_src, true) 
  end
  if plot == 'HistogramSeries' then
	ser = SgHistSeries:new(name, data_src, 1, true) 
  end
  if plot == 'HistogramSeriesBegin' then
	ser = SgHistSeries:new(name, data_src, 0, true) 
  end
  if plot == 'PieSeries' then
	ser = SgPieSeries:new(name, data_src, false, true) 
  end
  if plot == 'RoseSeries' then
	ser = SgRoseSeries:new(name, data_src, true) 
  end
  ser:setNameExt(title or SeriesList[cur_index].title)
  ser:setLevel(20)
  
  local ok = CfgFillStyle(ser, SeriesList[cur_index].id, order, SeriesList[cur_index].user)
  if (ok == nil) then
    CfgSetDefaultSeriesStyle(ser, plot)
  end

  setglobal(name, ser)
  return ser
end

-- Cria uma nova barra de erro associada a uma curva previamente
-- cadastrada
-- ser        : Serie
-- cur_index  : Indice do data_src em SeriesList ou nil caso a curva no
-- e1ds, e2ds : DataSrcs contendo a definicao dos offsets de erro
--
-- Retorna a barra de erro criada
function CfgBuildErrorBar(ser, cur_index, e1ds, e2ds)
  local name = CfgBuildName('errorbar'..SeriesList[cur_index].id..'_')
  local err  = SgErrorBar:new(name, ser, e1ds, e2ds)
  err:style():pen():setColor(ser:style():pen():color())
  err:setLevel(15)
  err:setNameExt('Barra de erro para '..ser:nameExt())
  setglobal(name, err)
  return err
end


-- Cria um novo eixo de acordo com os parametros da rea grfica
-- axe_type   : Tipo de eixo (0 = X, 1 = Y1, 2 = Y2 3 = Y3)
-- area_data  : Tabela contendo os dados da rea grfica
-- area_name  : Nome da rea grfica
-- serlist    : Tabela opcional onde so armazenadas referencias para todas as sries criadas
-- errorlist  : Tabela opcional onde so armazenadas referencias para todas as barras de erro criadas
--
-- Retorna o eixo criado.  Alem disso, para podermos aproveitar o esquema 
-- padro para salvar alteraes feitas pelo usurio, seta no ambiente global 
-- uma varivel de mesmo nome que o eixo referenciando-se a ele.
--
-- Se no houver especificacao de eixo para o axe_type recebido, retorna nil
-- 
function CfgBuildAxe(axe_type, area_data, area_name, serlist, errorlist)
  assert(axe_type >= 0 and axe_type <= 3)
  local index  = area_data[CfgGraphAreaFields.index[axe_type+1]]
  local series = area_data[CfgGraphAreaFields.series[axe_type+1]]
  local title  = area_data[CfgGraphAreaFields.title[axe_type+1]]

  if not index or index == 0 then return end

  -- Cria o eixo
  local name = CfgBuildName('axe'..AxeTypeList[index].id..'_')
  local unit = AxeTypeList[index].unit or ''
  local inv  = AxeTypeList[index].inv
  local kind = X_AXE
  local dir  = area_data.xaxeDir
  if axe_type > 0 then
    kind = Y_AXE
    if dir == HORIZONTAL then
      dir = VERTICAL
    else
      dir = HORIZONTAL
    end
  end
  local axe = SgAxe:new(name, kind, dir, inv, unit)
  setglobal(name, axe)
  axe:setNameExt(title)
  CfgSetAxeTag(axe, area_data.tag, axe_type)

  -- Carrega DataSrcs, cria e adiciona sries e barras de erro
  -- Para cada srie da tabela, cria e configura o DataSrc, lembrando
  -- que uma srie pode ser do tipo multidata onde cada entrada est
  -- associada a mltiplos datasrcs.
  if not series then return axe end

  for i = 1, getn(series) do 
    local v = series[i]
    local multi       = SeriesList[v].multidata
    local multiorder  = multi and 1  -- nil se no for multidata, 1 se for
    local single      = multi and SeriesList[v].singlestyle
    local singleStyle = nil
    local singleMarkerStyle = nil
    
    while 1 do
      local ds, e1, e2  = CfgLoadSeries(v, SeriesList[v].errorbar, SeriesList[v].user, 
                                        multiorder, area_name)
      if not ds then break end
      -- Se modo multiplo, descobre titulo da srie 
      local title = nil
      if multi and ds:name() ~= '' then
        title = ds:name()
      end
      -- Cria srie e adicona ao eixo
      local ser = CfgBuildSeries(ds, v, title, multiorder)
      CfgSetObjectTag(ser, area_data.tag, axe_type)
      axe:addSeries(ser)
      if serlist then
        tinsert(serlist, ser)
      end
      -- Cria barras de erro se apropriado
      if errorlist and e1 then
        local err = CfgBuildErrorBar(ser, v, e1, e2)
        CfgSetObjectTag(err, area_data.tag, axe_type)
        tinsert(errorlist, err)
      else
        tinsert(errorlist, nil)
      end
      -- Se a serie estiver operando em modo singlestyle (e esta no  a 
      -- primeira srie da sequencia), marca que esta srie no deve entrar
      -- na legenda e compartilha o estilo
      if single then
        if multiorder > 1 then
          ser:setShowInLegend(nil)
          if singleStyle       then ser:changeStyle(singleStyle) end
          if singleMarkerStyle then ser:changeMarkerStyle(singleMarkerStyle) end
        else
          if ser.changeStyle       then singleStyle = ser:style() end
          if ser.changeMarkerStyle then singleMarkerStyle = ser:markerStyle() end
        end
      end
      -- Se DataSrc  multiplo, repete para proximo numero de ordem
      if not multi then break end
      multiorder = multiorder + 1
    end
  end

  return axe
end


-- Adiciona barras de erro a uma rea grfica
--
-- area : Area grfica  qual as barras de erro pertencem
-- list : Lista de barras de erro disponiveis, conforme preenchido por CfgBuildAxe
-- axe  : Eixo Y ao qual as barras esto associadas
function CfgAddErrorBars(area, list, axe)
  for i = 1, getn(list) do
    if list[i] then
      area:addXYObject(list[i], axe)
    end
  end
end

-- Filtra listas de series e barras de erro para remover entradas
-- que nao devem aparecer no dialogo de configurao.
-- Sao filtradas aquelas curvas marcadas como singleStyle.
-- Estas entradas nao sao filtradas automaticamente em CfgBuildAxe pois
-- sao necessrias para a chamada a CfgAddErrorBars.
function CfgRemoveFilteredSeries(serlist, errlist)
  assert(getn(serlist) == getn(errlist))

  for i = getn(serlist), 1, -1 do
    if not serlist[i]:showInLegend() then
      tremove(serlist, i)
      tremove(errlist, i)
    end
  end
end


-- Cria e adiciona grid horizontal e vertical a uma rea
-- Adiciona tambm par de eixos regulares que permitem a configurao
-- do grfico com grid regular e no apenas com o grid associado a
-- um eixo.
--
-- area : Area grfica  qual os grids sero adicionados
-- xaxe : Eixo X ao qual o grid ser associado
-- yaxe : Eixo Y ao qual o grid ser associado
-- gridlist : List opcional a qual so acrescidos os grids criados (horizontal depois vertical)
function CfgAddGrids(area, xaxe, yaxe, gridlist)
  -- Cria grid associado ao eixo X
  local namex = CfgBuildName('grid_')
  local gridx = SgGrid:new(namex, xaxe, area)
  gridx:setNameExt(area:nameExt())
  area:addXObject(gridx)
  setglobal(namex, gridx)
  CfgSetObjectTag(gridx, CfgGetAreaTag(area), CfgGetAxeTag(xaxe))
  -- Cria grid associado ao eixo Y
  local namey = CfgBuildName('grid_')
  local gridy = SgGrid:new(namey, yaxe, area)
  gridy:setNameExt(area:nameExt())
  area:addYObject(gridy, yaxe)
  setglobal(namey, gridy)
  CfgSetObjectTag(gridy, CfgGetAreaTag(area), CfgGetAxeTag(yaxe))
  -- Adiciona grids a lista
  if gridlist then
    if xaxe:direction() == HORIZONTAL then
      tinsert(gridlist, gridy)
      tinsert(gridlist, gridx)
    else
      tinsert(gridlist, gridx)
      tinsert(gridlist, gridy)
    end
  end
  -- Cria eixos regulares e cadastra os mesmos no grafico
  local xregular_name = CfgBuildName('regularaxe_')
  local xregular = SgAxe:new(xregular_name, X_AXE, xaxe:direction(), nil, '')
  xregular:setLimits(0, 100)
  xregular:setStep(10, 2.5)
  xregular:setNameExt('Regular')
  setglobal(xregular_name, xregular)
  CfgSetAxeTag(xregular, CfgGetAreaTag(area), 4)

  local yregular_name = CfgBuildName('regularaxe_')
  local yregular = SgAxe:new(yregular_name, Y_AXE, yaxe:direction(), nil, '')
  yregular:setLimits(0, 100)
  yregular:setStep(10, 2.5)
  yregular:setNameExt('Regular')
  setglobal(yregular_name, yregular)
  CfgSetAxeTag(yregular, CfgGetAreaTag(area), 5)

  area:addRegularAxes(xregular, yregular, true)
end

-- 
-- Cria novo arquivo "file" contendo codigo Lua para recarregar
-- a tabela global Config
-- 
function CfgSaveState(file)
  assert(Config, "Tabela Config ausente")

  local f, err = openfile(file, "w")
  assert(f, err)

  CfgPersistTable = function(f, tab, level)
    local i, v = next(tab)
    while i do
      if type(i) == 'number' then
        write(f, strrep(' ', level*2), '[', i, '] = ')
      else
        write(f, strrep(' ', level*2), i, ' = ')
      end

      if type(v) == 'string' then
        write(f, format('%q', v))
      elseif type(v) == 'number' then
        write(f, v)
	  elseif type(v) == 'boolean' then
	    if v then
        write(f, 'true')
		else
		write(f, 'false')
		end
      elseif type(v) == 'nil' then
        write(f, 'nil')
      elseif type(v) == 'table' then
        write(f, '{\n')
        CfgPersistTable(f, v, level + 1)
        write(f, strrep(' ', level*2), '}')
      end
      write(f, ',\n')
      i, v = next(tab, i)
    end
  end

  write(f, "--\n")
  write(f, "-- SGD TemplateFile\n")
  write(f, "--\n\n")

  write(f, "-- Template configuration\n")
  write(f, "Config = {\n")
  CfgPersistTable(f, Config, 1)
  write(f, "}\n\n")

  
  write(f, "-- Template load\n")
  --write(f, format('loadtemplate(%q)\n', TemplatesList[Config.templateIndex].file))
  write(f, format('dofile(DefaultPath..%q)\n', TemplatesList[Config.templateIndex].file))
  write(f, "CurrentGraphic:clearChanges()\n\n")

  write(f, "-- User configurations\n")
  write(f, "if SkipUserConfiguration then return end\n")
  local userscript = CurrentGraphic:saveChangesToString()
  if userscript and userscript ~= '' then
    write(f, userscript)
  end

  closefile(f)
end


--
-- Executa script contendo configuracoes do usuario
-- Assume que o script foi gerado por uma chamada a funcao
-- saveChangesToString() do grafico.
-- Protege cada linha do script para o caso do objeto no 
-- existir mais
--
function CfgRunUserScript(script)
  script = gsub(script, '([^:\n]+)(.-)\n', 
                function(c1, c2)
                  if strfind(c1, '=') then  -- Objetos criados no script!!!
                    if strfind(c1, 'SgAxeMarker') then
                      local _,_,name,axe = strfind(c2,':new%("(.*)",.*,(.*),.*%)')
                      return 'if '..axe..' then '..c1..c2..' else '..name..' = nil end\n'
                    else
                      return c1..c2..'\n'
                    end
                  else 
                    local _,_,axe = strfind(c2, 'setAxe%((.*)%)')
                    if axe then -- alteracao do eixo de um grid tem de ser protegida
                      return 'if '..c1..' and '..axe..' then '..c1..c2..'end\n'
                    else
                      return 'if '..c1..' then '..c1..c2..' end\n'
                    end
                  end
                end
               )
--  local f, err = openfile('debug.txt', "w")
--  assert(f, err)
--  write(f, script)
--  closefile(f)
  if _VERSION == 'Lua 5.1' then
    dostring(script, "UserConfigScript")
  else
    assert(dostring(script, "UserConfigScript"))
  end
end

------------------------------------------------------------------
--
-- Funcoes para padronizacao de elementos dos templates
--
------------------------------------------------------------------

--
-- Cria guidelines para margens
-- Margens criadas so fixas a 5 mm da borda do grafico
-- Este tamanho pode ser alterado atravs da variavel global CfgMarginSize
-- Elementos sao publicados no ambiente global com os seguintes nomes:
--   glhBottomMargin
--   glhTopMargin
--   glvLeftMargin
--   glvRightMargin
-- Recebe como parmetro o grfico
--
CfgMarginSize = 5
function CfgBuildMarginGuidelines(g)
  glhBottomMargin = SgFixedGuideline:new(g, g:bottomGuideline(), CfgMarginSize)
  glhTopMargin    = SgFixedGuideline:new(g, g:topGuideline(),   -CfgMarginSize)
  glvLeftMargin   = SgFixedGuideline:new(g, g:leftGuideline(),   CfgMarginSize)
  glvRightMargin  = SgFixedGuideline:new(g, g:rightGuideline(), -CfgMarginSize)
end

--
-- Cria reas para ttulo e subttulo do grfico com nomes:
--    titleArea
--    subtitleArea
--
-- Cria Guidelines auxiliares para limitar a base das reas com nomes:
--    glhTitleBottom 
--    glhSubtitleBottom.
--
-- reas so inicilizadas de maneira visvel ou invisvel de 
-- acordo com a existncia ou no de ttulo para a rea
--
-- Assume que margens foram criadas respeitando os nomes utilizados
-- por CfgBuildMarginGuidelines()
--
function CfgBuildTitleAreas(g, title, subtitle)
  -- Cria linhas guia
  if title and title ~= '' then
    glhTitleBottom = SgLimitedGuideline:new(g, glhTopMargin, -10, -20, 100, 200)
  else
    glhTitleBottom = SgLimitedGuideline:new(g, glhTopMargin, 0, 0, 100, 200)
  end

  if subtitle and subtitle ~= '' then
    glhSubtitleBottom = SgLimitedGuideline:new(g, glhTitleBottom, -8, -15, 100, 200)
  else
    glhSubtitleBottom = SgLimitedGuideline:new(g, glhTitleBottom, 0, 0, 100, 200)
  end

  -- Cria e configura rea para ttulo
  titleArea = SgTextArea:new(g, "titleArea", glvLeftMargin, glvRightMargin, 
                             glhTitleBottom, glhTopMargin, title or '')
  titleArea:style():font():setColor(blue)
  titleArea:setTextSize(28, 12)
  if (not title) or title == '' then
    titleArea:setVisible(nil)
  end
  function titleArea:visibilityChanged(visible)
    if not visible then
      glhTitleBottom:setPos(0, 0, 100, 200)
    else
      glhTitleBottom:setPos(-10, -20, 100, 200)
    end
  end

  -- Cria e configura rea para subttulo
  subtitleArea = SgTextArea:new(g, "subtitleArea", glvLeftMargin, glvRightMargin, 
                                glhSubtitleBottom, glhTitleBottom, subtitle or '')
  subtitleArea:style():font():setColor(blue)
  subtitleArea:setTextSize(20, 10)
  if (not subtitle) or subtitle == '' then
    subtitleArea:setVisible(nil)
  end
  function subtitleArea:visibilityChanged(visible)
    if not visible then
      glhSubtitleBottom:setPos(0, 0, 100, 200)
    else
      glhSubtitleBottom:setPos(-8, -15, 100, 200)
    end
  end
end


--
-- Seta o estilo default para o fundo do grafico
--
function CfgSetGraphStyle(g)
end

--
-- Set a o estilo default para desenho do fundo e da borda de uma PlotArea
--
function CfgSetPlotAreaStyle(area)
  area:style():pen():setLineStyle(LS_CONTINUOUS)
  area:style():brush():setFillStyle(FS_SOLID)
  area:style():brush():setFgColor(gray)
end

--
-- Imprime dados basicos sobre todas as series criadas
--
function CfgDumpSeries()
  local g = globals()
  local i, v = next(g, nil)
  while i do
    if type(i) == 'string' and strfind(i, '^series') then
      print(format('name="%s", nameExt="%s", type=%d, showInLegend=%s', i, v:nameExt(), v:type(), tostring(v:showInLegend())))
    end
    i,v = next(g, i)
  end
end

